// Copyright 2014 Google Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package stats

import (
	"testing"
	"time"
)

func TestNanoTimer(t *testing.T) {
	s := &Stat{}
	func() {
		defer s.NanoTimer()()
		time.Sleep(time.Millisecond)
	}()
	if got := s.int64; got < 1000000 {
		t.Error("invalid nano time:", got)
	}
}
